@@ -1,12 +1,13 @@
 {-# LANGUAGE CPP #-}
 {-# LANGUAGE Rank2Types #-}
-{-# LANGUAGE FlexibleContexts #-}
 {-# LANGUAGE FlexibleInstances #-}
 {-# LANGUAGE UndecidableInstances #-}
 {-# LANGUAGE MultiParamTypeClasses #-}
 #if __GLASGOW_HASKELL__ >= 707
 {-# LANGUAGE DeriveDataTypeable #-}
+{-# LANGUAGE StandaloneDeriving #-}
 #endif
+#include "free-common.h"
 -----------------------------------------------------------------------------
 -- |
 -- Module      :  Control.Comonad.Cofree
@@ -33,6 +34,9 @@ module Control.Comonad.Cofree
   , _extract
   , _unwrap
   , telescoped
+  , telescoped_
+  , shoots
+  , leaves
   ) where
 
 import Control.Applicative
@@ -47,6 +51,7 @@ import Control.Category
 import Control.Monad(ap, (>=>), liftM)
 import Control.Monad.Zip
 import Data.Functor.Bind
+import Data.Functor.Classes.Compat
 import Data.Functor.Extend
 import Data.Data
 import Data.Distributive
@@ -56,7 +61,6 @@ import Data.Traversable
 import Data.Semigroup.Foldable
 import Data.Semigroup.Traversable
 import Prelude hiding (id,(.))
-import Prelude.Extras
 
 
 infixr 5 :<
@@ -99,11 +103,13 @@ infixr 5 :<
 --
 -- In particular, if @f a ≡ [a]@, the
 -- resulting data structure is a <https://en.wikipedia.org/wiki/Rose_tree Rose tree>.
--- For a practical application, check 
--- <https://personal.cis.strath.ac.uk/neil.ghani/papers/ghani-calco07 Higher Dimensional Trees, Algebraically> by Neil Ghani et al.
+-- For a practical application, check
+-- <https://web.archive.org/web/20161208002902/http://www.cs.le.ac.uk/people/ak155/Papers/CALCO-07/GK07.pdf Higher Dimensional Trees, Algebraically> by Neil Ghani et al.
 data Cofree f a = a :< f (Cofree f a)
 #if __GLASGOW_HASKELL__ >= 707
   deriving (Typeable)
+
+deriving instance (Typeable f, Data (f (Cofree f a)), Data a) => Data (Cofree f a)
 #endif
 
 -- | Use coiteration to generate a cofree comonad from a seed.
@@ -193,49 +199,90 @@ instance Alternative f => Applicative (Cofree f) where
   (<*>) = ap
   {-# INLINE (<*>) #-}
 
+#ifdef LIFTED_FUNCTOR_CLASSES
+instance (Show1 f) => Show1 (Cofree f) where
+  liftShowsPrec sp sl = go
+    where
+      goList = liftShowList sp sl
+      go d (a :< as) = showParen (d > 5) $
+        sp 6 a . showString " :< " . liftShowsPrec go goList 5 as
+#else
 instance (Functor f, Show1 f) => Show1 (Cofree f) where
   showsPrec1 d (a :< as) = showParen (d > 5) $
     showsPrec 6 a . showString " :< " . showsPrec1 5 (fmap Lift1 as)
+#endif
 
-instance (Show (f (Cofree f a)), Show a) => Show (Cofree f a) where
-  showsPrec d (a :< as) = showParen (d > 5) $
-    showsPrec 6 a . showString " :< " . showsPrec 5 as
+#ifdef LIFTED_FUNCTOR_CLASSES
+instance (Show1 f, Show a) => Show (Cofree f a) where
+#else
+instance (Functor f, Show1 f, Show a) => Show (Cofree f a) where
+#endif
+  showsPrec = showsPrec1
 
+#ifdef LIFTED_FUNCTOR_CLASSES
+instance (Read1 f) => Read1 (Cofree f) where
+  liftReadsPrec rp rl = go
+    where
+      goList = liftReadList rp rl
+      go d r = readParen (d > 5)
+        (\r' -> [(u :< v, w) |
+                (u, s) <- rp 6 r',
+                (":<", t) <- lex s,
+                (v, w) <- liftReadsPrec go goList 5 t]) r
+#else
 instance (Functor f, Read1 f) => Read1 (Cofree f) where
   readsPrec1 d r = readParen (d > 5)
                           (\r' -> [(u :< fmap lower1 v,w) |
                                   (u, s) <- readsPrec 6 r',
                                   (":<", t) <- lex s,
                                   (v, w) <- readsPrec1 5 t]) r
-
-instance (Read (f (Cofree f a)), Read a) => Read (Cofree f a) where
-  readsPrec d r = readParen (d > 5)
-                          (\r' -> [(u :< v,w) |
-                                  (u, s) <- readsPrec 6 r',
-                                  (":<", t) <- lex s,
-                                  (v, w) <- readsPrec 5 t]) r
-
-instance (Eq (f (Cofree f a)), Eq a) => Eq (Cofree f a) where
-#ifndef HLINT
-  a :< as == b :< bs = a == b && as == bs
 #endif
 
+#ifdef LIFTED_FUNCTOR_CLASSES
+instance (Read1 f, Read a) => Read (Cofree f a) where
+#else
+instance (Functor f, Read1 f, Read a) => Read (Cofree f a) where
+#endif
+  readsPrec = readsPrec1
+
+#ifdef LIFTED_FUNCTOR_CLASSES
+instance (Eq1 f, Eq a) => Eq (Cofree f a) where
+#else
+instance (Functor f, Eq1 f, Eq a) => Eq (Cofree f a) where
+#endif
+  (==) = eq1
+
+#ifdef LIFTED_FUNCTOR_CLASSES
+instance (Eq1 f) => Eq1 (Cofree f) where
+  liftEq eq = go
+    where
+      go (a :< as) (b :< bs) = eq a b && liftEq go as bs
+#else
 instance (Functor f, Eq1 f) => Eq1 (Cofree f) where
 #ifndef HLINT
-  a :< as ==# b :< bs = a == b && fmap Lift1 as ==# fmap Lift1 bs
+  eq1 (a :< as) (b :< bs) = a == b && eq1 (fmap Lift1 as) (fmap Lift1 bs)
+#endif
 #endif
 
-instance (Ord (f (Cofree f a)), Ord a) => Ord (Cofree f a) where
-  compare (a :< as) (b :< bs) = case compare a b of
-    LT -> LT
-    EQ -> compare as bs
-    GT -> GT
+#ifdef LIFTED_FUNCTOR_CLASSES
+instance (Ord1 f, Ord a) => Ord (Cofree f a) where
+#else
+instance (Functor f, Ord1 f, Ord a) => Ord (Cofree f a) where
+#endif
+  compare = compare1
 
+#ifdef LIFTED_FUNCTOR_CLASSES
+instance (Ord1 f) => Ord1 (Cofree f) where
+  liftCompare cmp = go
+    where
+      go (a :< as) (b :< bs) = cmp a b `mappend` liftCompare go as bs
+#else
 instance (Functor f, Ord1 f) => Ord1 (Cofree f) where
   compare1 (a :< as) (b :< bs) = case compare a b of
     LT -> LT
     EQ -> compare1 (fmap Lift1 as) (fmap Lift1 bs)
     GT -> GT
+#endif
 
 instance Foldable f => Foldable (Cofree f) where
   foldMap f = go where
@@ -366,3 +413,71 @@ telescoped :: Functor f =>
               (a -> f a) -> Cofree g a -> f (Cofree g a)
 telescoped = Prelude.foldr (\l r -> _unwrap . l . r) _extract
 {-# INLINE telescoped #-}
+
+-- not actually named 'eats'
+-- | Construct an @Lens@ into a @'Cofree' g@ given a list of lenses into the base functor.
+-- The only difference between this and 'telescoped' is that 'telescoped' focuses on a single value, but this focuses on the entire remaining subtree.
+-- When the input list is empty, this is equivalent to 'id'.
+-- When the input list is non-empty, this composes the input lenses
+-- with '_unwrap' to walk through the @'Cofree' g@.
+--
+-- For more on lenses see the 'lens' package on hackage.
+--
+-- @telescoped :: [Lens' (g ('Cofree' g a)) ('Cofree' g a)]      -> Lens' ('Cofree' g a) ('Cofree' g a)@
+--
+-- @telescoped :: [Traversal' (g ('Cofree' g a)) ('Cofree' g a)] -> Traversal' ('Cofree' g a) ('Cofree' g a)@
+--
+-- @telescoped :: [Getter (g ('Cofree' g a)) ('Cofree' g a)]     -> Getter ('Cofree' g a) ('Cofree' g a)@
+--
+-- @telescoped :: [Fold (g ('Cofree' g a)) ('Cofree' g a)]       -> Fold ('Cofree' g a) ('Cofree' g a)@
+--
+-- @telescoped :: [Setter' (g ('Cofree' g a)) ('Cofree' g a)]    -> Setter' ('Cofree' g a) ('Cofree' g a)@
+telescoped_ :: Functor f =>
+              [(Cofree g a -> f (Cofree g a)) -> g (Cofree g a) -> f (g (Cofree g a))] ->
+              (Cofree g a -> f (Cofree g a)) -> Cofree g a -> f (Cofree g a)
+telescoped_ = Prelude.foldr (\l r -> _unwrap . l . r) id
+{-# INLINE telescoped_ #-}
+
+-- | A @Traversal'@ that gives access to all non-leaf @a@ elements of a
+-- @'Cofree' g@ a, where non-leaf is defined as @x@ from @(x :< xs)@ where
+-- @null xs@ is @False@.
+--
+-- Because this doesn't give access to all values in the @'Cofree' g@,
+-- it cannot be used to change types.
+--
+-- @shoots :: Traversable g => Traversal' (Cofree g a) a@
+--
+-- N.B. On GHC < 7.9, this is slightly less flexible, as it has to
+-- use @null (toList xs)@ instead.
+shoots :: (Applicative f, Traversable g) => (a -> f a) -> Cofree g a -> f (Cofree g a)
+shoots f = go
+  where
+#if __GLASGOW_HASKELL__ < 709
+    go xxs@(x :< xs) | null (toList xs) = pure xxs
+#else
+    go xxs@(x :< xs) | null xs          = pure xxs
+#endif
+                     | otherwise        = (:<) <$> f x <*> traverse go xs
+{-# INLINE shoots #-}
+
+-- | A @Traversal'@ that gives access to all leaf @a@ elements of a
+-- @'Cofree' g@ a, where leaf is defined as @x@ from @(x :< xs)@ where
+-- @null xs@ is @True@.
+--
+-- Because this doesn't give access to all values in the @'Cofree' g@,
+-- it cannot be used to change types.
+--
+-- @shoots :: Traversable g => Traversal' (Cofree g a) a@
+--
+-- N.B. On GHC < 7.9, this is slightly less flexible, as it has to
+-- use @null (toList xs)@ instead.
+leaves :: (Applicative f, Traversable g) => (a -> f a) -> Cofree g a -> f (Cofree g a)
+leaves f = go
+  where
+#if __GLASGOW_HASKELL__ < 709
+    go (x :< xs) | null (toList xs) = (:< xs) <$> f x
+#else
+    go (x :< xs) | null xs          = (:< xs) <$> f x
+#endif
+                 | otherwise        = (x :<) <$> traverse go xs
+{-# INLINE leaves #-}
