@@ -1,5 +1,4 @@
 {-# LANGUAGE CPP #-}
-{-# LANGUAGE FlexibleContexts #-}
 {-# LANGUAGE FlexibleInstances #-}
 {-# LANGUAGE UndecidableInstances #-}
 {-# LANGUAGE MultiParamTypeClasses #-}
@@ -8,14 +7,8 @@
 #if __GLASGOW_HASKELL__ >= 707
 {-# LANGUAGE DeriveDataTypeable #-}
 #endif
+#include "free-common.h"
 
-#ifndef MIN_VERSION_base
-#define MIN_VERSION_base(x,y,z) 1
-#endif
-
-#ifndef MIN_VERSION_mtl
-#define MIN_VERSION_mtl(x,y,z) 1
-#endif
 -----------------------------------------------------------------------------
 -- |
 -- Module      :  Control.Monad.Trans.Free
@@ -42,6 +35,7 @@ module Control.Monad.Trans.Free
   , iterT
   , iterTM
   , hoistFreeT
+  , foldFreeT
   , transFreeT
   , joinFreeT
   , cutoff
@@ -59,9 +53,11 @@ module Control.Monad.Trans.Free
 
 import Control.Applicative
 import Control.Monad (liftM, MonadPlus(..), ap, join)
+import Control.Monad.Base (MonadBase(..))
 import Control.Monad.Catch (MonadThrow(..), MonadCatch(..))
 import Control.Monad.Trans.Class
 import Control.Monad.Free.Class
+import qualified Control.Monad.Fail as Fail
 import Control.Monad.IO.Class
 import Control.Monad.Reader.Class
 import Control.Monad.Writer.Class
@@ -69,15 +65,14 @@ import Control.Monad.State.Class
 import Control.Monad.Error.Class
 import Control.Monad.Cont.Class
 import Data.Functor.Bind hiding (join)
+import Data.Functor.Classes.Compat
 import Data.Monoid
-import Data.Function (on)
 import Data.Functor.Identity
 import Data.Traversable
 import Data.Bifunctor
 import Data.Bifoldable
 import Data.Bitraversable
 import Data.Data
-import Prelude.Extras
 
 #if !(MIN_VERSION_base(4,8,0))
 import Data.Foldable
@@ -91,15 +86,32 @@ data FreeF f a b = Pure a | Free (f b)
 #endif
            )
 
+#ifdef LIFTED_FUNCTOR_CLASSES
 instance Show1 f => Show2 (FreeF f) where
-  showsPrec2 d (Pure a)  = showParen (d > 10) $ showString "Pure " . showsPrec 11 a
-  showsPrec2 d (Free as) = showParen (d > 10) $ showString "Free " . showsPrec1 11 as
+  liftShowsPrec2 spa _sla _spb _slb d (Pure a) =
+    showsUnaryWith spa "Pure" d a 
+  liftShowsPrec2 _spa _sla spb slb d (Free as) =
+    showsUnaryWith (liftShowsPrec spb slb) "Free" d as
 
 instance (Show1 f, Show a) => Show1 (FreeF f a) where
-  showsPrec1 = showsPrec2
+  liftShowsPrec = liftShowsPrec2 showsPrec showList
+#else
+instance (Show1 f, Show a) => Show1 (FreeF f a) where
+  showsPrec1 d (Pure a)  = showParen (d > 10) $ showString "Pure " . showsPrec 11 a
+  showsPrec1 d (Free as) = showParen (d > 10) $ showString "Free " . showsPrec1 11 as
+#endif
 
+#ifdef LIFTED_FUNCTOR_CLASSES
 instance Read1 f => Read2 (FreeF f) where
-  readsPrec2 d r = readParen (d > 10)
+  liftReadsPrec2 rpa _rla rpb rlb = readsData $
+    readsUnaryWith rpa "Pure" Pure `mappend`
+    readsUnaryWith (liftReadsPrec rpb rlb) "Free" Free
+
+instance (Read1 f, Read a) => Read1 (FreeF f a) where
+  liftReadsPrec = liftReadsPrec2 readsPrec readList
+#else
+instance (Read1 f, Read a) => Read1 (FreeF f a) where
+  readsPrec1 d r = readParen (d > 10)
       (\r' -> [ (Pure m, t)
              | ("Pure", s) <- lex r'
              , (m, t) <- readsPrec 11 s]) r
@@ -107,26 +119,39 @@ instance Read1 f => Read2 (FreeF f) where
       (\r' -> [ (Free m, t)
              | ("Free", s) <- lex r'
              , (m, t) <- readsPrec1 11 s]) r
+#endif
 
-instance (Read1 f, Read a) => Read1 (FreeF f a) where
-  readsPrec1 = readsPrec2
-
+#ifdef LIFTED_FUNCTOR_CLASSES
 instance Eq1 f => Eq2 (FreeF f) where
-  Pure a  ==## Pure b = a == b
-  Free as ==## Free bs = as ==# bs
-  _       ==## _ = False
+  liftEq2 eq _ (Pure a) (Pure b) = eq a b
+  liftEq2 _ eq (Free as) (Free bs) = liftEq eq as bs
+  liftEq2 _ _ _ _ = False
 
 instance (Eq1 f, Eq a) => Eq1 (FreeF f a) where
-  (==#) = (==##)
+  liftEq = liftEq2 (==)
+#else
+instance (Eq1 f, Eq a) => Eq1 (FreeF f a) where
+  Pure a  `eq1` Pure b = a == b
+  Free as `eq1` Free bs = as `eq1` bs
+  _       `eq1` _ = False
+#endif
 
+#ifdef LIFTED_FUNCTOR_CLASSES
 instance Ord1 f => Ord2 (FreeF f) where
-  Pure a `compare2` Pure b = a `compare` b
-  Pure _ `compare2` Free _ = LT
-  Free _ `compare2` Pure _ = GT
-  Free fa `compare2` Free fb = fa `compare1` fb
+  liftCompare2 cmp _ (Pure a) (Pure b) = cmp a b
+  liftCompare2 _ _ (Pure _) (Free _) = LT
+  liftCompare2 _ _ (Free _) (Pure _) = GT
+  liftCompare2 _ cmp (Free fa) (Free fb) = liftCompare cmp fa fb
 
 instance (Ord1 f, Ord a) => Ord1 (FreeF f a) where
-  compare1 = compare2
+  liftCompare = liftCompare2 compare
+#else
+instance (Ord1 f, Ord a) => Ord1 (FreeF f a) where
+  Pure a `compare1` Pure b = a `compare` b
+  Pure _ `compare1` Free _ = LT
+  Free _ `compare1` Pure _ = GT
+  Free fa `compare1` Free fb = fa `compare1` fb
+#endif
 
 instance Functor f => Functor (FreeF f a) where
   fmap _ (Pure a)  = Pure a
@@ -179,31 +204,81 @@ free :: FreeF f a (Free f a) -> Free f a
 free = FreeT . Identity
 {-# INLINE free #-}
 
-deriving instance Eq (m (FreeF f a (FreeT f m a))) => Eq (FreeT f m a)
+#ifdef LIFTED_FUNCTOR_CLASSES
+instance (Eq1 f, Eq1 m, Eq a) => Eq (FreeT f m a) where
+#else
+instance (Functor f, Eq1 f, Functor m, Eq1 m, Eq a)=> Eq (FreeT f m a) where
+#endif
+    (==) = eq1
 
+#ifdef LIFTED_FUNCTOR_CLASSES
+instance (Eq1 f, Eq1 m) => Eq1 (FreeT f m) where
+  liftEq eq = go
+    where
+      go (FreeT x) (FreeT y) = liftEq (liftEq2 eq go) x y
+#else
 instance (Functor f, Eq1 f, Functor m, Eq1 m) => Eq1 (FreeT f m) where
-  (==#) = on (==#) (fmap (Lift1 . fmap Lift1) . runFreeT)
+  eq1 = on eq1 (fmap (Lift1 . fmap Lift1) . runFreeT)
+#endif
 
-deriving instance Ord (m (FreeF f a (FreeT f m a))) => Ord (FreeT f m a)
+#ifdef LIFTED_FUNCTOR_CLASSES
+instance (Ord1 f, Ord1 m, Ord a) => Ord (FreeT f m a) where
+#else
+instance (Functor f, Ord1 f, Functor m, Ord1 m, Ord a) => Ord (FreeT f m a) where
+#endif
+    compare = compare1
 
+#ifdef LIFTED_FUNCTOR_CLASSES
+instance (Ord1 f, Ord1 m) => Ord1 (FreeT f m) where
+  liftCompare cmp = go
+    where
+      go (FreeT x) (FreeT y) = liftCompare (liftCompare2 cmp go) x y
+#else
 instance (Functor f, Ord1 f, Functor m, Ord1 m) => Ord1 (FreeT f m) where
   compare1 = on compare1 (fmap (Lift1 . fmap Lift1) . runFreeT)
+#endif
 
+#ifdef LIFTED_FUNCTOR_CLASSES
+instance (Show1 f, Show1 m) => Show1 (FreeT f m) where
+  liftShowsPrec sp sl = go
+    where
+      goList = liftShowList sp sl
+      go d (FreeT x) = showsUnaryWith
+        (liftShowsPrec (liftShowsPrec2 sp sl go goList) (liftShowList2 sp sl go goList))
+        "FreeT" d x
+#else
 instance (Functor f, Show1 f, Functor m, Show1 m) => Show1 (FreeT f m) where
   showsPrec1 d (FreeT m) = showParen (d > 10) $
     showString "FreeT " . showsPrec1 11 (Lift1 . fmap Lift1 <$> m)
+#endif
 
-instance Show (m (FreeF f a (FreeT f m a))) => Show (FreeT f m a) where
-  showsPrec d (FreeT m) = showParen (d > 10) $
-    showString "FreeT " . showsPrec 11 m
+#ifdef LIFTED_FUNCTOR_CLASSES
+instance (Show1 f, Show1 m, Show a) => Show (FreeT f m a) where
+#else
+instance (Functor f, Show1 f, Functor m, Show1 m, Show a) => Show (FreeT f m a) where
+#endif
+  showsPrec = showsPrec1
 
+#ifdef LIFTED_FUNCTOR_CLASSES
+instance (Read1 f, Read1 m) => Read1 (FreeT f m) where
+  liftReadsPrec rp rl = go
+    where
+      goList = liftReadList rp rl
+      go = readsData $ readsUnaryWith
+        (liftReadsPrec (liftReadsPrec2 rp rl go goList) (liftReadList2 rp rl go goList))
+        "FreeT" FreeT
+#else
 instance (Functor f, Read1 f, Functor m, Read1 m) => Read1 (FreeT f m) where
   readsPrec1 d =  readParen (d > 10) $ \r ->
     [ (FreeT (fmap lower1 . lower1 <$> m),t) | ("FreeT",s) <- lex r, (m,t) <- readsPrec1 11 s]
+#endif
 
-instance Read (m (FreeF f a (FreeT f m a))) => Read (FreeT f m a) where
-  readsPrec d =  readParen (d > 10) $ \r ->
-    [ (FreeT m,t) | ("FreeT",s) <- lex r, (m,t) <- readsPrec 11 s]
+#ifdef LIFTED_FUNCTOR_CLASSES
+instance (Read1 f, Read1 m, Read a) => Read (FreeT f m a) where
+#else
+instance (Functor f, Read1 f, Functor m, Read1 m, Read a) => Read (FreeT f m a) where
+#endif
+  readsPrec = readsPrec1
 
 instance (Functor f, Monad m) => Functor (FreeT f m) where
   fmap f (FreeT m) = FreeT (liftM f' m) where
@@ -223,13 +298,17 @@ instance (Functor f, Monad m) => Bind (FreeT f m) where
   (>>-) = (>>=)
 
 instance (Functor f, Monad m) => Monad (FreeT f m) where
-  fail e = FreeT (fail e)
   return = pure
   {-# INLINE return #-}
   FreeT m >>= f = FreeT $ m >>= \v -> case v of
     Pure a -> runFreeT (f a)
     Free w -> return (Free (fmap (>>= f) w))
 
+  fail = Fail.fail
+
+instance (Functor f, Monad m) => Fail.MonadFail (FreeT f m) where
+  fail e = FreeT (fail e)
+
 instance MonadTrans (FreeT f) where
   lift = FreeT . liftM Pure
   {-# INLINE lift #-}
@@ -238,6 +317,10 @@ instance (Functor f, MonadIO m) => MonadIO (FreeT f m) where
   liftIO = lift . liftIO
   {-# INLINE liftIO #-}
 
+instance (Functor f, MonadBase b m) => MonadBase b (FreeT f m) where
+  liftBase = lift . liftBase
+  {-# INLINE liftBase #-}
+
 instance (Functor f, MonadReader r m) => MonadReader r (FreeT f m) where
   ask = lift ask
   {-# INLINE ask #-}
@@ -332,6 +415,15 @@ instance (Monad m, Traversable m, Traversable f) => Traversable (FreeT f m) wher
 hoistFreeT :: (Monad m, Functor f) => (forall a. m a -> n a) -> FreeT f m b -> FreeT f n b
 hoistFreeT mh = FreeT . mh . liftM (fmap (hoistFreeT mh)) . runFreeT
 
+-- | The very definition of a free monad transformer is that given a natural
+-- transformation you get a monad transformer homomorphism.
+foldFreeT :: (MonadTrans t, Monad (t m), Monad m)
+          => (forall n x. Monad n => f x -> t n x) -> FreeT f m a -> t m a
+foldFreeT f (FreeT m) = lift m >>= foldFreeF
+  where
+    foldFreeF (Pure a) = return a
+    foldFreeF (Free as) = f as >>= foldFreeT f
+
 -- | Lift a natural transformation from @f@ to @g@ into a monad homomorphism from @'FreeT' f m@ to @'FreeT' g m@
 transFreeT :: (Monad m, Functor g) => (forall a. f a -> g a) -> FreeT f m b -> FreeT g m b
 transFreeT nt = FreeT . liftM (fmap (transFreeT nt) . transFreeF nt) . runFreeT
