@@ -1,18 +1,10 @@
 {-# LANGUAGE CPP #-}
-{-# LANGUAGE FlexibleContexts #-}
 {-# LANGUAGE FlexibleInstances #-}
 {-# LANGUAGE UndecidableInstances #-}
 {-# LANGUAGE MultiParamTypeClasses #-}
 {-# LANGUAGE Rank2Types #-}
 {-# LANGUAGE DeriveDataTypeable #-}
-
-#ifndef MIN_VERSION_base
-#define MIN_VERSION_base(x,y,z) 1
-#endif
-
-#ifndef MIN_VERSION_mtl
-#define MIN_VERSION_mtl(x,y,z) 1
-#endif
+#include "free-common.h"
 
 -----------------------------------------------------------------------------
 -- |
@@ -81,6 +73,7 @@ import Control.Monad.Catch (MonadCatch(..), MonadThrow(..))
 import Control.Monad (ap, liftM, MonadPlus(..), join)
 import Control.Monad.Fix
 import Control.Monad.Trans.Class
+import qualified Control.Monad.Fail as Fail
 import Control.Monad.Free.Class
 import Control.Monad.State.Class
 import Control.Monad.Error.Class
@@ -92,14 +85,13 @@ import Data.Bifunctor
 import Data.Bitraversable
 import Data.Either
 import Data.Functor.Bind hiding (join)
+import Data.Functor.Classes.Compat
 import Data.Functor.Identity
-import Data.Function (on)
-import Data.Monoid
+import Data.Semigroup
 import Data.Semigroup.Foldable
 import Data.Semigroup.Traversable
 import Data.Typeable
 import Data.Data
-import Prelude.Extras
 
 #if !(MIN_VERSION_base(4,8,0))
 import Data.Foldable hiding (fold)
@@ -133,33 +125,81 @@ runIter :: Iter a -> Either a (Iter a)
 runIter = runIdentity . runIterT
 {-# INLINE runIter #-}
 
+#ifdef LIFTED_FUNCTOR_CLASSES
+instance (Eq1 m) => Eq1 (IterT m) where
+  liftEq eq = go
+    where
+      go (IterT x) (IterT y) = liftEq (liftEq2 eq go) x y
+#else
 instance (Functor m, Eq1 m) => Eq1 (IterT m) where
-  (==#) = on (==#) (fmap (fmap Lift1) . runIterT)
+  eq1 = on eq1 (fmap (fmap Lift1) . runIterT)
+#endif
 
-instance Eq (m (Either a (IterT m a))) => Eq (IterT m a) where
-  IterT m == IterT n = m == n
+#ifdef LIFTED_FUNCTOR_CLASSES
+instance (Eq1 m, Eq a) => Eq (IterT m a) where
+#else
+instance (Functor m, Eq1 m, Eq a) => Eq (IterT m a) where
+#endif
+  (==) = eq1
 
+#ifdef LIFTED_FUNCTOR_CLASSES
+instance (Ord1 m) => Ord1 (IterT m) where
+  liftCompare cmp = go
+    where
+      go (IterT x) (IterT y) = liftCompare (liftCompare2 cmp go) x y
+#else
 instance (Functor m, Ord1 m) => Ord1 (IterT m) where
   compare1 = on compare1 (fmap (fmap Lift1) . runIterT)
+#endif
 
-instance Ord (m (Either a (IterT m a))) => Ord (IterT m a) where
-  compare (IterT m) (IterT n) = compare m n
+#ifdef LIFTED_FUNCTOR_CLASSES
+instance (Ord1 m, Ord a) => Ord (IterT m a) where
+#else
+instance (Functor m, Ord1 m, Ord a) => Ord (IterT m a) where
+#endif
+  compare = compare1
 
+#ifdef LIFTED_FUNCTOR_CLASSES
+instance (Show1 m) => Show1 (IterT m) where
+  liftShowsPrec sp sl = go
+    where
+      goList = liftShowList sp sl
+      go d (IterT x) = showsUnaryWith
+        (liftShowsPrec (liftShowsPrec2 sp sl go goList) (liftShowList2 sp sl go goList))
+        "IterT" d x     
+#else
 instance (Functor m, Show1 m) => Show1 (IterT m) where
   showsPrec1 d (IterT m) = showParen (d > 10) $
     showString "IterT " . showsPrec1 11 (fmap (fmap Lift1) m)
+#endif
 
-instance Show (m (Either a (IterT m a))) => Show (IterT m a) where
-  showsPrec d (IterT m) = showParen (d > 10) $
-    showString "IterT " . showsPrec 11 m
+#ifdef LIFTED_FUNCTOR_CLASSES
+instance (Show1 m, Show a) => Show (IterT m a) where
+#else
+instance (Functor m, Show1 m, Show a) => Show (IterT m a) where
+#endif
+  showsPrec = showsPrec1
 
+#ifdef LIFTED_FUNCTOR_CLASSES
+instance (Read1 m) => Read1 (IterT m) where
+  liftReadsPrec rp rl = go
+    where
+      goList = liftReadList rp rl
+      go = readsData $ readsUnaryWith
+        (liftReadsPrec (liftReadsPrec2 rp rl go goList) (liftReadList2 rp rl go goList))
+        "IterT" IterT
+#else
 instance (Functor m, Read1 m) => Read1 (IterT m) where
   readsPrec1 d =  readParen (d > 10) $ \r ->
     [ (IterT (fmap (fmap lower1) m),t) | ("IterT",s) <- lex r, (m,t) <- readsPrec1 11 s]
+#endif
 
-instance Read (m (Either a (IterT m a))) => Read (IterT m a) where
-  readsPrec d =  readParen (d > 10) $ \r ->
-    [ (IterT m,t) | ("IterT",s) <- lex r, (m,t) <- readsPrec 11 s]
+#ifdef LIFTED_FUNCTOR_CLASSES
+instance (Read1 m, Read a) => Read (IterT m a) where
+#else
+instance (Functor m, Read1 m, Read a) => Read (IterT m a) where
+#endif
+  readsPrec = readsPrec1
 
 instance Monad m => Functor (IterT m) where
   fmap f = IterT . liftM (bimap f (fmap f)) . runIterT
@@ -176,6 +216,10 @@ instance Monad m => Monad (IterT m) where
   {-# INLINE return #-}
   IterT m >>= k = IterT $ m >>= either (runIterT . k) (return . Right . (>>= k))
   {-# INLINE (>>=) #-}
+  fail = Fail.fail
+  {-# INLINE fail #-}
+
+instance Monad m => Fail.MonadFail (IterT m) where
   fail _ = never
   {-# INLINE fail #-}
 
@@ -240,7 +284,7 @@ instance MonadWriter w m => MonadWriter w (IterT m) where
   listen (IterT m) = IterT $ liftM concat' $ listen (fmap listen `liftM` m)
     where
       concat' (Left  x, w) = Left (x, w)
-      concat' (Right y, w) = Right $ second (w <>) <$> y
+      concat' (Right y, w) = Right $ second (w `mappend`) <$> y
   pass m = IterT . pass' . runIterT . hoistIterT clean $ listen m
     where
       clean = pass . liftM (\x -> (x, const mempty))
@@ -388,17 +432,9 @@ interleave_ [] = return ()
 interleave_ xs = IterT $ liftM (Right . interleave_ . rights) $ mapM runIterT xs
 {-# INLINE interleave_ #-}
 
-instance (Monad m, Monoid a) => Monoid (IterT m a) where
+instance (Monad m, Semigroup a, Monoid a) => Monoid (IterT m a) where
   mempty = return mempty
-  x `mappend` y = IterT $ do
-    x' <- runIterT x
-    y' <- runIterT y
-    case (x', y') of
-      ( Left a, Left b)  -> return . Left  $ a `mappend` b
-      ( Left a, Right b) -> return . Right $ liftM (a `mappend`) b
-      (Right a, Left b)  -> return . Right $ liftM (`mappend` b) a
-      (Right a, Right b) -> return . Right $ a `mappend` b
-
+  mappend = (<>)
   mconcat = mconcat' . map Right
     where
       mconcat' :: (Monad m, Monoid a) => [Either a (IterT m a)] -> IterT m a
@@ -416,7 +452,17 @@ instance (Monad m, Monoid a) => Monoid (IterT m a) where
 
       compact' a []               = [Left a]
       compact' a (r@(Right _):xs) = (Left a):(r:(compact xs))
-      compact' a (  (Left a'):xs) = compact' (a <> a') xs
+      compact' a (  (Left a'):xs) = compact' (a `mappend` a') xs
+
+instance (Monad m, Semigroup a) => Semigroup (IterT m a) where
+  x <> y = IterT $ do
+    x' <- runIterT x
+    y' <- runIterT y
+    case (x', y') of
+      ( Left a, Left b)  -> return . Left  $ a <> b
+      ( Left a, Right b) -> return . Right $ liftM (a <>) b
+      (Right a, Left b)  -> return . Right $ liftM (<> b) a
+      (Right a, Right b) -> return . Right $ a <> b
 
 #if __GLASGOW_HASKELL__ < 707
 instance Typeable1 m => Typeable1 (IterT m) where
